/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#include <string.h>
#include <sstream>
#include <iostream>
#include <map>
#include <ctime>
#include <assert.h>
#include <errno.h>
#include <stdlib.h>

#ifndef _WIN32
# include <sys/poll.h>
#endif

#include <DxcCallback.hpp>
#include <Connection/DxcConnectorImp.hpp>
#include <Parser/MessageDriver.hpp>

using namespace Dxc;
using namespace std;

#ifdef _WIN32

#pragma warning (disable:4996)

typedef int socklen_t;

int initSock()
{
    WSADATA wsd;
    WORD version = MAKEWORD(2, 2);

    int err = WSAStartup(version, &wsd);
    if (err != 0)
    {
        string e("WSAStartup failed with error: ");
        e += err;
        throw DxcException(e);
    }

    return err;
}

bool recvBufferTooSmall()
{
    if (WSAGetLastError() == WSAEMSGSIZE)
        return true;
    else
        return false;
}

bool DxcConnectorImp::socketWriteReady(int sockfd)
{
    fd_set fds;

    fds.fd_count = 1;
    fds.fd_array[0] = sockfd;

    timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 300000;

    int stat = select(1, NULL, &fds, NULL, &tv);

    if (stat == 1)
        return true;

    return false;
}

bool DxcConnectorImp::socketReadReady(int sockfd)
{
    fd_set fds;

    FD_ZERO(&fds);
    FD_SET(sockfd, &fds);

    timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 5000;

    int stat = select(1, &fds, NULL, NULL, &tv);

    if (stat == 1)
        return true;

    return false;
}

#else

#include <arpa/inet.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netdb.h>

int initSock()
{
    return 0;
}

bool recvBufferTooSmall()
{
    /*
        if (WSAGetLastError() == WSAEMSGSIZE)
            return true;
        else
            return false;
    */
    return false;
}

bool DxcConnectorImp::socketWriteReady(int sockfd)
{
    pollfd pfd;
    pfd.fd = sockfd;
    pfd.events = POLLOUT;
    int stat = poll(&pfd,1,50);

    if (stat > 0)
    {
        short r = pfd.revents;
        if (r & POLLHUP)
        {
            cout << "poll_socket(): poll() detected hangup (POLLHUP)" << endl;
            return false;
        }
        else if (r & POLLERR)
        {
            cout << "poll_socket(): poll() POLLERR" << endl;
            return false;
        }
        else if (r & POLLNVAL)
        {
            cout <<"poll_socket(): poll() POLLNVAL" << endl;
            return false;
        }
        else
        {
            return true;
        }
    }
}

bool DxcConnectorImp::socketReadReady(int sockfd)
{
    pollfd pfd;
    pfd.fd = sockfd;
    pfd.events = POLLIN | POLLPRI;
    int stat = poll(&pfd,1,50);

    if (stat > 0)
    {
        short r = pfd.revents;
        if (r & POLLHUP)
        {
            cout << "poll_socket(): poll() detected hangup (POLLHUP)" << endl;
            return false;
        }
        else if (r & POLLERR)
        {
            cout << "poll_socket(): poll() POLLERR" << endl;
            return false;
        }
        else if (r & POLLNVAL)
        {
            cout <<"poll_socket(): poll() POLLNVAL" << endl;
            return false;
        }
        else
        {
            return true;
        }
    }
}
#endif

DxcConnectorImp::DxcConnectorImp(DxcCallback * cb, const string& logFilePrefix)
        : callback_(cb), logFile_(NULL), shutdown_(false)
{
    initSock();

    if (logFilePrefix != "")
    {
        time_t rawtime;
        tm* t;
        time(&rawtime);
        t = localtime(&rawtime);

        ostringstream nameStr;
        nameStr << logFilePrefix;
        nameStr << "-" << (1900 + t->tm_year)
        << "-" << t->tm_mon << "-" << t->tm_mday << "-"
        << t->tm_hour << t->tm_min << ".log";

        logFile_ = new ofstream(nameStr.str().c_str());
        *logFile_ << "foo;" << endl;
        logFile_->close();
    }
}

void DxcConnectorImp::acceptConnection(int port, int *sockp)
{
    if ((listenfd_ = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
        ostringstream e;
        e << "Couldn't connect on port " << port;
        throw DxcException(e.str());
    }
#ifdef _WIN32
    const char on = 1;
#else
    int on = 1;
#endif
    if (setsockopt(listenfd_, SOL_SOCKET, SO_REUSEADDR, &on,  sizeof(on)))
    {
        ostringstream e;
        e << "Couldn't set bind socket options. Socket: " << listenfd_;
        throw DxcException(e.str());
    }

    struct sockaddr_in serv;

    memset(&serv, 0, sizeof(struct sockaddr_in));

    serv.sin_family = AF_INET;
    serv.sin_addr.s_addr = htonl(INADDR_ANY);
    serv.sin_port = htons(port);

    if (bind(listenfd_, (struct sockaddr *)&serv, sizeof(serv)) < 0)
    {
        ostringstream e;
        e << "Couldn't bind on port " << port;
        throw DxcException(e.str());
    }
    if (listen(listenfd_, SOMAXCONN) < 0)
    {
        ostringstream e;
        e << "Couldn't listen on port " << port;
        throw DxcException(e.str());
    }

    cout << "Listening on port: " << port << endl;

    ReceiveParam * rp = new ReceiveParam;
    rp->sockfdp = sockp;
    *(rp->sockfdp) = listenfd_;
    rp->cb = callback_;
    rp->shutdown = &shutdown_;

    pthread_t * recvThread = new pthread_t;

    int threadRet = pthread_create(recvThread, NULL, DxcConnectorImp::acceptAndReceive, rp);

    if (threadRet)
    {
        delete recvThread, rp;
        throw DxcException("Unable to create recvThread.");
    }

    recvThreads_.push_back(recvThread);
}

int DxcConnectorImp::connect(string host, int port, int *sockp)
{
    int tmpsock;

    if ((tmpsock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
    {
        string e("Couldn't create socket.");
        throw DxcException(e);
    }

    struct hostent *h = gethostbyname(host.c_str());
    if (NULL == h)
    {
        string e("Couldn't find hostname.");
        throw DxcException(e);
    }

    struct sockaddr_in client;

    memset(&client, 0, sizeof(client));

    client.sin_family = AF_INET;
    client.sin_addr.s_addr = *((unsigned long int *)(h->h_addr));
    client.sin_port = htons(port);

    if (::connect(tmpsock, (struct sockaddr *)&client, sizeof(client)) < 0)
    {
        ostringstream e;
        e << "Couldn't connect on port " << port;
        throw DxcException(e.str());
    }

    *sockp = tmpsock;

    ReceiveParam * rp = new ReceiveParam;
    rp->sockfdp = sockp;
    rp->cb = callback_;
    rp->shutdown = &shutdown_;

    pthread_t * recvThread = new pthread_t;

    int threadRet = pthread_create(recvThread, NULL, DxcConnectorImp::receive, rp);

    if (threadRet)
    {
        string e("Couldn't create recvThread.");
        throw DxcException(e);
    }

    recvThreads_.push_back(recvThread);

    return *sockp;
}

void DxcConnectorImp::transmit(int sockfd, const DxcData *msg)
{
    if (sockfd < 0)
    {
        return;
    }
    else if (!socketWriteReady(sockfd))
    {
        cerr << "Couldn't send message. Socket: " << sockfd << endl;
        return;
    }

    ostringstream s;
    s << msg;
    string msgStr(s.str());

    if (send(sockfd, msgStr.c_str(), msgStr.length(), NULL) < 0)
    {
        cerr << "Couldn't send message. Socket: " << sockfd << endl;
    }

    if (logFile_ && logFile_->is_open())
    {
        *logFile_ << msgStr << endl;
        logFile_->flush();
    }
}

void *DxcConnectorImp::receive(void * receiveParam)
{
    ReceiveParam * rp = static_cast<ReceiveParam *>(receiveParam);

    if (rp->cb == NULL || *(rp->sockfdp) < 0)
    {
        delete rp;
        return NULL;
    }

    DxcCallback* callback = rp->cb;
    string cbtype(typeid(*callback).name());
    int sockfd = *(rp->sockfdp);

    size_t len = 4096;
    char* buf = new char[len];
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = 500;

    int readRet = 1;
    MessageDriver md;
    string strBuf = "";

    if (setsockopt(sockfd, SOL_SOCKET, SO_RCVTIMEO, (char *)&tv,  sizeof tv))
    {
        cout << "receive() couldn't set socket options. Socket: " << sockfd << endl;
        return NULL;
    }

    while ( !(*(rp->shutdown)) )
    {
        if (socketReadReady(sockfd))
        {
            readRet = recv(sockfd, buf, len, 0);
            if (recvBufferTooSmall())
            {
                delete buf;
                len *= 2;
                buf = new char[len];
            }
            else if (readRet > 0)
            {
                string::size_type n;
                strBuf.append(buf, readRet);

                n = strBuf.find_last_of(';');

                if (n < strBuf.length())
                {
                    string tmpStr(strBuf, 0, n + 1);
                    strBuf.erase(0, n + 1);

                    try
                    {
                        md.parseString(tmpStr);
                    }
                    catch (DxcException ex)
                    {
                        ostringstream err;
                        err << ex.what() << ". Attempting to parse: " << tmpStr;
                        cout << err.str();

                        ErrorData ed(DxcData::currentTime(), err.str());
                        callback->onDataReceived(&ed);
                    }

                    if (md.hasMessages())
                    {
                        const DxcData* m = NULL;
                        while (m = md.getNextMessage())
                        {
                            if (callback)
                                callback->onDataReceived(m);
                            delete m;
                        }
                    }
                }
            }
        }
    }

    closeSocket(sockfd);

    delete buf;
    return NULL;
}

void *DxcConnectorImp::acceptAndReceive(void * receiveParam)
{
    ReceiveParam * rp = static_cast<ReceiveParam *>(receiveParam);

    if (rp->cb == NULL || *(rp->sockfdp) < 0)
    {
        delete rp;
        return NULL;
    }

    int sockfd = *(rp->sockfdp);
    *(rp->sockfdp) = -1; // so that external thread doesn't send/rcv until
    // the connection is established
    struct sockaddr_in child;

    memset(&child, 0, sizeof(child));
    socklen_t childlen = (socklen_t)sizeof(child);
    if ((sockfd = accept(sockfd, (struct sockaddr *)&child, &childlen)) < 0)
    {
        cout << "Couldn't accept connection. Socket: " << sockfd;
        return NULL;
    }

    (*rp->sockfdp) = sockfd;

    receive(rp);
    return NULL;
}

DxcConnectorImp::~DxcConnectorImp()
{
    shutdown_ = true;

    // join threads and delete pointers in recvThreads_
    while (!recvThreads_.empty())
    {
        pthread_t * tp = recvThreads_.back();
        pthread_join(*tp, NULL);
        delete tp;
        recvThreads_.pop_back();
    }

    if (logFile_ && logFile_->is_open())
        logFile_->close();

    closeSocket(listenfd_);
}

void DxcConnectorImp::closeSocket(int socket)
{
    int ret = shutdown(socket, 1);
#ifdef _WIN32
    closesocket(socket);

#pragma warning (default:4996)
#else
    close(socket);
#endif
}
