/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#include <DataTypes/DiagnosisData.hpp>

using namespace std;
using namespace Dxc;

DiagnosisData::DiagnosisData(Timestamp timestamp,
                             bool detectionSignal,
                             bool isolationSignal,
                             const CandidateSet& isolation,
                             const std::string& notes):
        DxcData(timestamp),
        detectionSignal_(detectionSignal),
        isolationSignal_(isolationSignal),
        candidateSet_(isolation),
        notes_(notes)
{
}

DiagnosisData::DiagnosisData(bool detectionSignal,
                             bool isolationSignal,
                             const CandidateSet& isolation,
                             const std::string& notes):
        DxcData(),
        detectionSignal_(detectionSignal),
        isolationSignal_(isolationSignal),
        candidateSet_(isolation),
        notes_(notes)
{
}

DiagnosisData::DiagnosisData(Timestamp timestamp,
                             bool detectionSignal,
                             bool isolationSignal,
                             const vector<vector<Literal *> *>& candidates,
                             const vector<double>& weights,
                             const std::string& notes):
        DxcData(timestamp),
        detectionSignal_(detectionSignal),
        isolationSignal_(isolationSignal),
        candidateSet_(),
        notes_(notes)
{
    vector<vector<Literal *> *>::const_iterator ci;
    vector<Literal *>::const_iterator fi;
    vector<double> tmpWeights(weights);

    for (ci = candidates.begin(); ci != candidates.end(); ci++)
    {
        map<string, string> faultMap;
        vector<Literal *> * faultlist = *ci;

        for (fi = faultlist->begin(); fi != faultlist->end(); fi++)
        {
            Literal * fault = *fi;
            const Value * val = (fault->getValue());
            const StringValue * strFault = dynamic_cast<const StringValue*>(val);
            if (strFault)
                faultMap[*(fault->getId())] = strFault->get();
        }

        Candidate c;
        c.faultMap = faultMap;
        if (!weights.empty())
        {
            c.weight = tmpWeights.front();
            tmpWeights.erase(tmpWeights.begin());
        }

        candidateSet_.insert(c);
    }
}

bool DiagnosisData::ltCandidate::operator()(const DiagnosisData::Candidate& c1, const DiagnosisData::Candidate& c2)
{
    map<string, string>::const_iterator i1 = c1.faultMap.begin(), i2 = c2.faultMap.begin();

    if (i1 == c1.faultMap.end())
        return true;

    while (!(i1 == c1.faultMap.end() || i2 == c2.faultMap.end()))
    {
        if (i1->first < i2->first)
        {
            return true;
        }
        else if (i1->first > i2->first)
        {
            return false;
        }
        else
        {
            if (i1->second < i2->second)
            {
                return true;
            }
            else if (i1->second > i2->second)
            {
                return false;
            }
            else
            {
                i1++;
                i2++;
            }
        }
    }

    return false;
}

DiagnosisData* DiagnosisData::clone() const
{
    return new DiagnosisData(*this);
}

ostream &DiagnosisData::put(ostream &out) const
{
    out << "diagnosis @" << this->getTimeStamp() << " ";
    out << "detection = "
    << (this->getDetectionSignal() ? "true" : "false");
    out << ", isolation = "
    << (this->getIsolationSignal() ? "true" : "false");
    out << ", candidates = {";

    DiagnosisData::CandidateSet cs = this->getCandidateSet();
    DiagnosisData::CandidateSet::iterator i = cs.begin();

    for (; i != cs.end(); i++)
    {
        if (i != cs.begin())
            out << ", ";

        out << "{";

        map<string, string>::const_iterator j = i->faultMap.begin();

        for (; j != i->faultMap.end(); j++)
        {
            if (j != i->faultMap.begin())
                out << ", ";

            out << j->first << " = " << j->second;
        }

        out << " }";
    }

    out << "}, weights = {";

    for (i = cs.begin(); i != cs.end(); i++)
    {
        if (i != cs.begin())
            out << ", ";

        out << i->weight;
    }

    out << "}, notes = '"
    << this->getNotes() << "';" << std::endl;

    out.flush();

    return out;
}

bool DiagnosisData::getDetectionSignal() const
{
    return detectionSignal_;
}

bool DiagnosisData::getIsolationSignal() const
{
    return isolationSignal_;
}

DiagnosisData::CandidateSet DiagnosisData::getCandidateSet() const
{
    return candidateSet_;
}

string DiagnosisData::getNotes() const
{
    return notes_;
}
