/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#include <fstream>
#include <sstream>

#include <DataTypes/DxcException.hpp>

#include "MessageDriver.hpp"
#include "MessageScanner.hpp"

using namespace Dxc;
using namespace std;

MessageDriver::MessageDriver() : traceScanning_(false),
        traceParsing_(false),
        messageHook_(NULL)
{
    pthread_mutex_init(&mQueueMutex_, NULL);
}

void MessageDriver::gotMessage(const DxcData *msg)
{
    if (NULL != messageHook_)
    {
        messageHook_(msg);
    }

    pthread_mutex_lock(&mQueueMutex_);
    mQueue_.push_back(msg);
    pthread_mutex_unlock(&mQueueMutex_);
}

const DxcData* MessageDriver::getNextMessage()
{
    if (mQueue_.empty())
        return NULL;

    const DxcData* m = mQueue_.front();
    mQueue_.pop_front();
    return m;
}

bool MessageDriver::hasMessages()
{
    return !(mQueue_.empty());
}

bool MessageDriver::parseStream(istream &in, const string &streamName)
{
    streamName_ = streamName;

    MessageScanner scanner(&in);
    MessageParser parser(*this);

    scanner.set_debug(traceScanning_);
    parser.set_debug_level(traceParsing_);

    this->lexer_ = &scanner;

    return parser.parse() == 0;
}

bool MessageDriver::parseString(const string &input, const string &sname)
{
    istringstream iss(input);

    return parseStream(iss, sname);
}

bool MessageDriver::parseFile(const string &fileName)
{
    ifstream inputStream(fileName.c_str());
    if (!inputStream.good())
    {
        throw DxcException("Error opening '" + fileName + "'.");
    }
    return parseStream(inputStream, fileName);
}

void MessageDriver::error(const class location &l,
                          const string &m)
{
    ostringstream es;
    es << l << ": " << m;
    throw DxcException(es.str());
}

void MessageDriver::error(const string &m)
{
    throw DxcException(m);
}

void MessageDriver::setMessageHook(void (*messageHook)(const DxcData *))
{
    messageHook_ = messageHook;
}
