package org.dxc.api.datatypes;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;

import org.dxc.api.DxcConfig;
import org.w3c.dom.*;

/**
 * System catalog parser.
 * 
 * Provides a data structure for accessing information contained in a system's
 * XML catalog.
 */
public class SystemCatalog {
    private static Map<String, SystemCatalog> catMap;
    private Map<String, String> compIdToType;
    private Map<String, Set<String>> typeToModes;

    private SystemCatalog(String system) throws Exception {
        compIdToType = new HashMap<String, String>();
        typeToModes = new HashMap<String, Set<String>>();
        String sysDir = DxcConfig.getInstance().get("SYSTEMCAT_DIR");
        File f = new File(sysDir + system + ".xml");

        DocumentBuilder db = DocumentBuilderFactory.newInstance()
                .newDocumentBuilder();
        Document doc = db.parse(f);
        doc.getDocumentElement().normalize();

        NodeList nl = doc.getElementsByTagName("component");

        // get component ids

        for (int i = 0; i < nl.getLength(); i++) {
            Node nm = (Node) nl.item(i);
            String componentID = getChildByName(nm, "name").getFirstChild()
                    .getNodeValue();
            String componentType = getChildByName(nm, "componentType")
                    .getFirstChild().getNodeValue();
            compIdToType.put(componentID, componentType);
        }

        // get component type -> modes map

        Map<String, String> typeMdRefMap = new HashMap<String, String>();
        Map<String, Set<String>> mdRefModesMap = new HashMap<String, Set<String>>();

        nl = doc.getElementsByTagName("componentTypeCatalog").item(0)
                .getChildNodes();

        for (int i = 0; i < nl.getLength(); i++) {
            Node nm = (Node) nl.item(i);
            if (nm.getNodeType() == Node.ELEMENT_NODE) {
                String componentType = getChildByName(nm, "name")
                        .getFirstChild().getNodeValue();

                String modesRef = getChildByName(nm, "modesRef")
                        .getFirstChild().getNodeValue();

                typeMdRefMap.put(componentType, modesRef);
            }
        }

        nl = doc.getElementsByTagName("modeGroup");

        for (int i = 0; i < nl.getLength(); i++) {
            Node nm = (Node) nl.item(i);
            if (nm.getNodeType() == Node.ELEMENT_NODE) {
                Node mgNode = getChildByName(nm, "name");
                String modeGroup = mgNode.getFirstChild().getNodeValue();
                Set<String> modes = new HashSet<String>();

                Node md = mgNode.getNextSibling();

                while (md != null) {
                    if (md.getNodeName().equals("mode")) {
                        NamedNodeMap att = md.getAttributes();
                        Node typeN = att.getNamedItem("xsi:type");
                        String type = typeN.getNodeValue();

                        if (type.equals("faultMode")) {
                            modes.add(getChildByName(md, "name")
                                    .getFirstChild().getNodeValue());
                        }
                    }
                    md = md.getNextSibling();
                }

                mdRefModesMap.put(modeGroup, modes);
            }

            Iterator<String> j = typeMdRefMap.keySet().iterator();
            while (j.hasNext()) {
                String type = j.next();
                String modeRef = typeMdRefMap.get(type);
                Set<String> modes = mdRefModesMap.get(modeRef);

                typeToModes.put(type, modes);
            }
        }
    }

    /**
     * Get a system's catalog.
     * 
     * @param system
     *            the system's name
     * 
     *            System catalogs are searched for in $DXC_HOME/[SYSTEMCAT_DIR]
     *            (see DXC.cfg for SYSTEMCAT_DIR).
     */
    public static SystemCatalog getInstance(String system) throws Exception {
        if (catMap == null) {
            catMap = new HashMap<String, SystemCatalog>();
        }

        if (catMap.containsKey(system)) {
            return catMap.get(system);
        } else {
            SystemCatalog s = new SystemCatalog(system);
            catMap.put(system, s);
            return s;
        }
    }

    /**
     * The set of component IDs the system contains.
     */
    public Set<String> getComponentIDs() {
        return compIdToType.keySet();
    }

    /**
     * The set of mode IDs of a given component
     * 
     * @param componentID
     *            component's ID
     */
    public Set<String> getFaultModes(String componentID) {
        String type = compIdToType.get(componentID);

        if (type == null)
            return null;

        return typeToModes.get(type);
    }

    /**
     * Test the validity of a componentID.
     * 
     * @param componentID
     *            component's ID
     */
    public boolean isComponentValid(String componentID) {
        if (compIdToType.keySet().contains(componentID))
            return true;

        return false;
    }

    /**
     * Given a valid component, test the validity of a mode ID.
     * 
     * @param componentID
     *            component ID
     * @param modeID
     *            mode ID
     */
    public boolean isFaultModeValid(String componentID, String modeID) {
        if (compIdToType.keySet().contains(componentID)) {
            String type = compIdToType.get(componentID);
            if (this.typeToModes.get(type).contains(modeID))
                return true;
        }

        return false;
    }

    private Node getChildByName(Node n, String name) {
        NodeList nl = n.getChildNodes();
        for (int i = 0; i < nl.getLength(); i++) {
            Node ret = nl.item(i);
            if (ret.getNodeName().equals(name))
                return ret;
        }
        return null;
    }
}
