/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#include <stdlib.h>
#endif

#include <iostream>

#include <DxcDataTypes.hpp>
#include <DataTypes/SystemCatalog.hpp>
#include <Connection/ConnectorFactory.hpp>

#include "Parser/MessageDriver.hpp"

#ifdef _WIN32
#include <Pthreads-w32/pthread.h>
#define pthreadsInit pthread_win32_process_attach_np
#define pthreadsStop pthread_win32_process_detach_np
#else
#include <pthread.h>
#define pthreadsInit void
#define pthreadsStop void
#endif

using namespace Dxc;
using namespace std;

static DxcConnector *connector;
static pthread_mutex_t connectorMutex;
static pthread_cond_t daReady;
static const SystemCatalog* sd;

class SDSCallback : public DxcCallback
{
private:
    virtual void processData(const DxcData *msg)
    {
        if (string(typeid(*msg).name()) == string(typeid(ScenarioStatusData).name()))
        {
            const ScenarioStatusData* stat = dynamic_cast<const ScenarioStatusData *>(msg);

            if (stat->getStatus() == ScenarioStatusData::SR_READY)
            {
                ScenarioStatusData ssd(ScenarioStatusData::SDS_AWAITING_DA);
                pthread_mutex_lock(&connectorMutex);
                connector->sendMessage(&ssd);
                pthread_mutex_unlock(&connectorMutex);
            }
            else if (stat->getStatus() == ScenarioStatusData::DA_READY)
            {
                ScenarioStatusData ssd(ScenarioStatusData::STARTED);
                pthread_mutex_lock(&connectorMutex);
                connector->sendMessage(&ssd);
                pthread_mutex_unlock(&connectorMutex);

                pthread_cond_broadcast(&daReady);
            }
        }
        else if (string(typeid(*msg).name()) == string(typeid(ErrorData).name()))
        {
            pthread_mutex_lock(&connectorMutex);
            if (connector)
            {
                connector->sendMessage(msg);
                ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);
                connector->sendMessage(&ssd);
                exit(0);
            }
            pthread_mutex_unlock(&connectorMutex);
        }
    };
};

void validateData(const DxcData* msg)
{
    if (!sd)
        return;

    if ((string)typeid(*msg).name() == (string)typeid(SensorData).name())
    {
        const SensorData* s = dynamic_cast<const SensorData*>(msg);
        SensorData::SensorValueMap svm = s->getSensorValueMap();;
        SensorData::SensorValueMap::iterator i;
        i = svm.begin();

        for (; i != svm.end(); i++)
        {
            if (!sd->isComponentValid(i->first))
            {
                ostringstream oss;
                oss << "Invalid component ID: " << i->first <<
                " at timestamp " << s->getTimeStamp();
                throw DxcException(oss.str());
            }
        }
    }
    if ((string)typeid(*msg).name() == (string)typeid(FaultInjectData).name())
    {
        const FaultInjectData* fid = dynamic_cast<const FaultInjectData*>(msg);
        string componentID = fid->getComponentID();
        string modeID = fid->getModeID();

        if (!sd->isComponentValid(componentID))
        {
            ostringstream oss;
            oss << "Invalid component ID: " << componentID <<
            " in faultInjection at timestamp " << fid->getTimeStamp();
            throw DxcException(oss.str());
        }

        if (!sd->isFaultModeValid(componentID, modeID))
        {
            ostringstream oss;
            oss << "Invalid mode ID: " << modeID <<
            " in faultInjection at timestamp " << fid->getTimeStamp();
            throw DxcException(oss.str());
        }
    }
};

void sendMessage(const DxcData *msg)
{
    static Timestamp startTime = -1;
    static Timestamp msgTime0 = msg->getTimeStamp();
    static bool once = true;

    if (once)
    {
        startTime = DxcData::currentTime();

        once = false;
    }

    validateData(msg);

    Timestamp msgTime = msg->getTimeStamp();
    Timestamp msgTimeNew = startTime + (msgTime - msgTime0);

    long sleepTime = (long)(msgTimeNew - DxcData::currentTime());

    DxcData* newMsg = msg->clone();
    newMsg->setTimeStamp(msgTimeNew);

    if (sleepTime > 0)
        msleep(sleepTime);

    pthread_mutex_lock(&connectorMutex);
    DxcConnector * tmp = connector;
    if (connector)
        connector->sendMessage(newMsg);
    pthread_mutex_unlock(&connectorMutex);

    delete newMsg;
}

void removeQuotes(string& str)
{
    size_t n = str.find_first_of("'\"");
    while (n != string::npos)
    {
        str.erase(n, 1);
        n = str.find_first_of("'\"", n);
    }
}


int main(int argc, char **argv)
{
    try
    {
        SDSCallback sdsc;
        pthread_mutex_t condMutex;

        pthreadsInit();
        pthread_mutex_init(&connectorMutex, NULL);
        pthread_mutex_init(&condMutex, NULL);
        pthread_cond_init(&daReady, NULL);


        connector = ConnectorFactory::getSDSConnector(&sdsc);

        MessageDriver *scenarioDriver = new MessageDriver();
        scenarioDriver->setMessageHook(sendMessage);

        pthread_mutex_lock(&condMutex);
        pthread_cond_wait(&daReady, &condMutex);
        pthread_mutex_unlock(&condMutex);

        if (argc != 3)
        {
            for (int i = 0; i < argc; i++)
                cout << argv[i] << endl;
            throw DxcException("Usage: " + (string)argv[0] + " [system_name] [scenario_file]");
        }
        else
        {
            sd = SystemCatalog::getInstance(argv[1]);
            cout << "Parsing scenario file: " << endl << argv[2] << endl;
            cout.flush();
            string s(argv[2]);
            removeQuotes(s);
            scenarioDriver->parseFile(s);
        }
        delete scenarioDriver;

        pthread_mutex_lock(&connectorMutex);
        ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);
        if (connector)
        {
            connector->sendMessage(&ssd);
        }
        delete connector;
        connector = NULL;
        pthread_mutex_unlock(&connectorMutex);

    }
    catch (DxcException ex)
    {
        try
        {
            cout << ex.what() << endl;

            pthread_mutex_lock(&connectorMutex);
            if (connector)
            {
                ErrorData ed(DxcData::currentTime(), ex.what());
                ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);
                connector->sendMessage(&ed);
                connector->sendMessage(&ssd);
                delete connector;
                connector = NULL;
            }
            pthread_mutex_unlock(&connectorMutex);
        }
        catch (...)
        {
            cerr << "SDS: Unable to send error. " << endl;
        }
    }
    catch (...)
    {
        cerr << "SDS: Unknown exception." << endl;
    }

    return 0;
}
