/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include <iostream>

#include <Connection/ConnectorFactory.hpp>
#include <DataTypes/SystemCatalog.hpp>
#include <DxcDataTypes.hpp>

#include "Parser/MessageDriver.hpp"

#ifdef _WIN32
#include <Pthreads-w32/pthread.h>
#define pthreadsInit pthread_win32_process_attach_np
#define pthreadsStop pthread_win32_process_detach_np
#else
#include <pthread.h>
#define pthreadsInit void
#define pthreadsStop void
#endif

using namespace Dxc;
using namespace std;

static ostream* output = NULL;
static DxcConnector *connector = NULL;
static pthread_mutex_t connectorMutex;
static pthread_cond_t daReady;
static const SystemCatalog* sd;

class SDSCallback : public DxcCallback
{
private:
    virtual void processData(const DxcData *msg)
    {
        if (string(typeid(*msg).name()) == string(typeid(ScenarioStatusData).name()))
        {
            const ScenarioStatusData* stat = dynamic_cast<const ScenarioStatusData *>(msg);

            if (stat->getStatus() == ScenarioStatusData::DA_READY)
            {
                ScenarioStatusData ssd(ScenarioStatusData::STARTED);
                pthread_mutex_lock(&connectorMutex);
                connector->sendMessage(&ssd);
                pthread_mutex_unlock(&connectorMutex);

                pthread_cond_broadcast(&daReady);
            }
        }

        *output << msg;
    };
};

void validateData(const DxcData* msg)
{
    if (!sd)
        return;

    if ((string)typeid(*msg).name() == (string)typeid(SensorData).name())
    {
        const SensorData* s = dynamic_cast<const SensorData*>(msg);
        SensorData::SensorValueMap svm = s->getSensorValueMap();;
        SensorData::SensorValueMap::iterator i;
        i = svm.begin();

        for (; i != svm.end(); i++)
        {
            if (!sd->isComponentValid(i->first))
            {
                ostringstream oss;
                oss << "Invalid component ID: " << i->first <<
                " at timestamp " << s->getTimeStamp();
                throw DxcException(oss.str());
            }
        }
    }
    if ((string)typeid(*msg).name() == (string)typeid(FaultInjectData).name())
    {
        const FaultInjectData* fid = dynamic_cast<const FaultInjectData*>(msg);
        string componentID = fid->getComponentID();
        string modeID = fid->getModeID();

        if (!sd->isComponentValid(componentID))
        {
            ostringstream oss;
            oss << "Invalid component ID: " << componentID <<
            " in faultInjection at timestamp " << fid->getTimeStamp();
            throw DxcException(oss.str());
        }

        if (!sd->isFaultModeValid(componentID, modeID))
        {
            ostringstream oss;
            oss << "Invalid mode ID: " << modeID <<
            " in faultInjection at timestamp " << fid->getTimeStamp();
            throw DxcException(oss.str());
        }
    }
};

void sendMessage(const DxcData *msg)
{
    static Timestamp startTime = -1;
    static Timestamp msgTime0 = msg->getTimeStamp();
    static bool once = true;

    if (once)
    {
        startTime = DxcData::currentTime();

        once = false;
    }

    validateData(msg);

    Timestamp msgTime = msg->getTimeStamp();
    Timestamp msgTimeNew = startTime + (msgTime - msgTime0);

    long sleepTime = (long)(msgTimeNew - DxcData::currentTime());

    DxcData* newMsg = msg->clone();
    newMsg->setTimeStamp(msgTimeNew);

    if (sleepTime > 0)
        msleep(sleepTime);

    pthread_mutex_lock(&connectorMutex);
    DxcConnector * tmp = connector;
    if (connector)
        connector->sendMessage(newMsg);
    pthread_mutex_unlock(&connectorMutex);

    delete newMsg;
}

void printUsage(const string& cmd)
{
    cout << "usage: " << cmd << " [-h] [-v system_name] [-t wait_time]";
    cout << " input_file [output_file]" << endl;
    cout << "       [-h]:             displays this message" << endl;
    cout << "       [-v system_name]: validates scenario file against specified" << endl
         << "                         system's catalog." << endl;
    cout << "       [-t wait_time]:   specifies how long StandaloneSDS should" << endl
         << "                         wait after scenario end before exiting." << endl;
    cout << "       input_file:       scenario to run (must be in proper format)." << endl;
    cout << "       [output_file]:    optional file to store results (otherwise" << endl
         << "                         use stdout)." << endl << endl;
}

int main(int argc, char **argv)
{
    output = &cout;
    pthreadsInit();
    pthread_mutex_init(&connectorMutex, NULL);
    pthread_cond_init(&daReady, NULL);

    try
    {
        long waitTime = 0;
        if (argc < 2)
        {
            printUsage(argv[0]);
            throw DxcException("Too few arguments, exiting.");
        }

        int pos = 1;

        while (pos < argc && argv[pos][0] == '-')
        {
            switch (argv[pos][1])
            {
            case 'h':
                printUsage(argv[0]);
                pos++;
                break;
            case 'v':
                sd = SystemCatalog::getInstance(argv[++pos]);
                pos++;
                break;
            case 't':
            {
                waitTime = -1;
                istringstream iss(argv[++pos]);
                try
                {
                    iss >> waitTime;
                    pos++;
                }
                catch (...)
                {
                    throw DxcException("-t: unable to parse time.");
                }
                if (waitTime == -1)
                    throw DxcException("-t: unable to parse time.");
                break;
            }
            default:
                throw DxcException(string(argv[pos]) + ": unknown option.");
            }
        }

        if (pos >= argc)
        {
            printUsage(argv[0]);
            throw DxcException("No input file specified.");
        }

        ifstream tmpStream(argv[pos]);
        if (!tmpStream.good())
            throw DxcException(string(argv[pos]) +
                               ": unable to open file");
        tmpStream.close();

        if (pos + 1 < argc)
        {
            try
            {
                cout << "Writing to: " << endl << argv[pos + 1] << endl;
                output = new ofstream(argv[pos + 1]);
                if (!output->good())
                    throw DxcException("");
            }
            catch (...)
            {
                cout << "Unable to open results file for writing. Writing to stdout." << endl;
                output = &cout;
            }
        }

        // now start the scenario.
        SDSCallback sdsc;
        pthread_mutex_t condMutex;
        pthread_mutex_init(&condMutex, NULL);

        DxcConnector* slc = ConnectorFactory::getScenarioLoaderConnector(NULL);
        msleep(500);
        connector = ConnectorFactory::getSDSConnector(&sdsc);
        msleep(500);
        DxcConnector* src = ConnectorFactory::getScenarioRecorderConnector(&sdsc);

        MessageDriver *scenarioDriver = new MessageDriver();
        scenarioDriver->setMessageHook(sendMessage);


        pthread_mutex_lock(&condMutex);
        cout << "Awaiting DA..." << endl;
        pthread_cond_wait(&daReady, &condMutex);
        pthread_mutex_unlock(&condMutex);

        msleep(500);
        cout << "Parsing scenario file: " << endl << argv[pos] << endl;
        scenarioDriver->parseFile(argv[pos]);

        delete scenarioDriver;

        pthread_mutex_lock(&connectorMutex);
        ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);
        connector->sendMessage(&ssd);

        if (waitTime > 0)
            msleep(waitTime*1000);
        delete connector, slc, src;
        connector = NULL;
        pthread_mutex_unlock(&connectorMutex);

    }
    catch (DxcException ex)
    {
        try
        {
            cout << ex.what() << endl;

            pthread_mutex_lock(&connectorMutex);
            if (connector)
            {
                ErrorData ed(DxcData::currentTime(), ex.what());
                ScenarioStatusData ssd(ScenarioStatusData::SDS_ENDED);
                connector->sendMessage(&ed);
                connector->sendMessage(&ssd);
                delete connector;
                connector = NULL;
            }
            pthread_mutex_unlock(&connectorMutex);
        }
        catch (...)
        {
            cerr << "SDS: Unable to send error. " << endl;
        }
    }
    catch (...)
    {
        cerr << "SDS: Unknown exception." << endl;
    }

    return 0;
}
