/*
  DXC Framework (http://dx-competition.org/)
  Copyright 2008 David Garcia

  The DXC Framework is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as published by the
  Free Software Foundation, either version 3 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
  details.

  You should have received a copy of the GNU Lesser General Public License
  along with this program. If not, see <http://www.gnu.org/licenses/>
*/
#include <iostream>
#include <exception>
#include <typeinfo>
#include <algorithm>

#include <DxcDataTypes.hpp>
#include <Connection/ConnectorFactory.hpp>

#include <DAConfig.hpp>
#include <ScenarioLoader.hpp>
#include <ProcessFactory.hpp>
#include <FileLister.hpp>
#include <DxcConfig.hpp>

#ifdef _WIN32
#include <Windows.h>
#include <Pthreads-w32/pthread.h>
#define pthreadsInit pthread_win32_process_attach_np
#define pthreadsStop pthread_win32_process_detach_np
#else
#include <unistd.h>
#include <pthread.h>
#define pthreadsInit void
#define pthreadsStop void
#endif

using namespace std;
using namespace Dxc;

vector<DAConfig> getDAConfigs(const vector<string>& daDirs)
{
    vector<DAConfig> ret;
    vector<string>::const_iterator i = daDirs.begin();
    string algsDir;

    try
    {
        DxcConfig* cfg = DxcConfig::getInstance();
        algsDir = cfg->get("ALGS_DIR");

        for (; i != daDirs.end(); i++)
        {
            ret.push_back(DAConfig(algsDir + *i + "config.xml"));
        }
    }
    catch (DxcException ex)
    {
        cout << "ScenarioLoader unable to load config for DA in "
             << algsDir << *i << endl;
    }

    return ret;
}

void runScenario(const DAConfig& daConf, const string& scenarioFile, const string& systemName)
{
    ScenarioTracker* strack(0);
    SLCallback* slcb(0);
    DxcConnector* slc(0);
    ProcessLoader *sds(0), *sr(0), *da(0);

    try
    {
        DxcConfig* cfg = DxcConfig::getInstance();

        string scenDir = cfg->get("SCENARIOS_DIR");
        string binDir = cfg->get("BIN_DIR");
        string resultsDir = cfg->get("RESULTS_DIR");

        string scenario = "\"" + scenDir +
                          systemName + DIR_DELIM +
                          scenarioFile + "\"";

        string results = "\"" + resultsDir +
                         daConf.getName() +
                         " - " + systemName +
                         " - " + scenarioFile + "\"";

        sds =
            ProcessFactory::createProcessLoader(
                "SDS",
                binDir,
                systemName + " " + scenario);

        sr =
            ProcessFactory::createProcessLoader(
                "ScenarioRecorder",
                binDir,
                results);

        cout << "Loading Diagnoser: " << daConf.getName() << endl;
        da = ProcessFactory::createDAProcessLoader(daConf, systemName);

        strack = new ScenarioTracker();
        slcb = new SLCallback(strack);
        slc = ConnectorFactory::getScenarioLoaderConnector(slcb);

        sds->start();
        msleep(500);
        sr->start();
        int stat = ScenarioTracker::ST_INIT;

        if (strack->getStatus() != ScenarioTracker::ST_AWAITING_DA)
        {
            stat = strack->waitForUpdate(cfg->getLong("SDS_SR_TIMEOUT_MS"));
            if (stat != ScenarioTracker::ST_AWAITING_DA &&
                    strack->getStatus() != ScenarioTracker::ST_AWAITING_DA)
            {
                throw DxcException("Timed out waiting for SDS and Recorder.");
            }
        }

        slc->initConnections();

        da->start();

        if (strack->getStatus() != ScenarioTracker::ST_STARTED)
        {
            stat = strack->waitForUpdate(cfg->getLong("DA_TIMEOUT_MS"));
            if (stat != ScenarioTracker::ST_STARTED &&
                    strack->getStatus() != ScenarioTracker::ST_STARTED)
            {
                throw DxcException(
                    "Timed out waiting for scenario start (check DiagnosisAlgorithm->SDS connection).");
            }
        }

        stat = strack->waitForUpdate(cfg->getLong("SCN_END_TIMEOUT_MS"));
        if (stat != ScenarioTracker::ST_ENDED)
        {
            if (stat == ScenarioTracker::ST_ERROR)
            {
                throw DxcException(
                    "Received error message. ScenarioLoader Exiting.");
            }

            throw DxcException("Timed out waiting for scenario end.");
        }

        if (!da->timedWait(cfg->getLong("POST_SCN_TIME_MS")))
        {
            cout << "DA has not exited. Terminating." << endl;
            da->terminate();
        }

        ScenarioStatusData ssd(ScenarioStatusData::DA_ENDED);
        slc->sendMessage(&ssd);

        ProfilingData profilingData(da->getCpuTime(),
                                    da->getPeakMemoryUsage());
        slc->sendMessage(&profilingData);

        delete slc;
        delete slcb;
        delete sds;
        delete sr;
        delete da;
        delete strack;
    }
    catch (DxcException ex)
    {
        string err("SL Error: ");
        err += ex.what();
        cerr << err << endl;

        ErrorData ed(DxcData::currentTime(), err);
        if (slc)
        {
            slc->sendMessage(&ed);
        }

        delete slc;
        delete slcb;
        // wait for apps to exit gracefully
        msleep(3000);
        delete sds;
        delete sr;
        delete da;
        delete strack;
    }
}

void runExperiment(const DAConfig& daConf)
{
    vector<string> systems = daConf.getSupportedSystems();
    vector<string>::const_iterator i;

    DxcConfig* cfg = DxcConfig::getInstance();
    string scenDir = cfg->get("SCENARIOS_DIR");

    for (i = systems.begin(); i != systems.end(); i++)
    {
        vector<string> scenarios = FileLister::getFiles(scenDir + *i + DIR_DELIM);

        vector<string>::const_iterator j = scenarios.begin();

        for (; j  != scenarios.end(); j++)
        {
            runScenario(daConf, *j, *i);
            msleep(3000);
        }
    }
}

int main(int argc, char* argv[])
{
    pthreadsInit();

    try
    {
        DxcConfig* cfg = DxcConfig::getInstance();
        string algsDir = cfg->get("ALGS_DIR");

        vector<string> daDirs = FileLister::getDirectories(algsDir);
        vector<DAConfig> daCfgs = getDAConfigs(daDirs);

        for_each(daCfgs.begin(), daCfgs.end(), runExperiment);

        cout << "Exiting..." << endl;
        msleep(2000);
    }
    catch (DxcException ex)
    {
        cerr << "Error: " << ex.what() << endl;
    }

    pthreadsStop();
}
